/****************************************************************************

	PSP NAND flash installer for devhook 

****************************************************************************/

#include "common.h"

#define USE_MALLOC 0

/*
when "flash1:/registry/system.ireg","flash1:/registry/system.dreg" are removed and flash1 full,
FW1.50 fails in the initialization of the registry file and repeats this forever.
FW1.50 doesn't delete other flash1 file in registry initialization.
Therefore, flash1 always needs a some free area for safety.

BTW.
If the FAT area of flash1 crashed, PSP may not work too.
So I thinks that the recovery mode like CUSTOM FIRMWARE of Dark Alex is necessary for FW1.50 safety.
*/
#define FLASH1_KEEP_FREE_SIZE (1024*1024)

/****************************************************************************
	works
****************************************************************************/
static char src_path[128];
static char dst_path[128];

/****************************************************************************
	current firmware path
****************************************************************************/

// "ms0:/dh/[FW]"
static const char *ms_cur_path(void)
{
	return dh_config.flash;
}

// "/dh/[FW]"
static const char *flash_cur_path(void)
{
	char *ptr =strchr(dh_config.flash,':');
	if(ptr) return ptr+1;
	return dh_config.flash;
}

/****************************************************************************
	make directry
****************************************************************************/
static int make_directry(const char *path)
{
	char tmp[128];
	char *ptr;
	int result = 0;

	strcpy(tmp,path);
	ptr = strchr(tmp,':');
	if(ptr) ptr++;
	if(ptr[0]=='/') ptr++;

	while( (ptr = strchr(ptr,'/') ) != NULL)
	{
		*ptr = 0x00;
		printf("Mkdir %s:",tmp);
		result = sceIoMkdir(tmp,0x777);
#if 0
		if( result < 0 )
			printf("Err %08X",result);
		else
			printf("OK");
#endif
		printf("\n");
		*ptr = '/';
		ptr++;
	}
	return result;
}
/****************************************************************************
	copy/remove file one
****************************************************************************/
static unsigned char file_buf[0x4000];
static int buf_size = 0x4000;

static int copy_file(const char *src,const char *dst)
{
	SceUID fds;
	SceUID fdd;
	int result = 0;
	int readed;
	if(dst)
	{
		printf("Copy %s:",src);

		fds = sceIoOpen(src, PSP_O_RDONLY, 0777);
		if(fds < 0 )
		{
			printf("open error\n");
			return -1;
		}
		fdd = sceIoOpen(dst, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);
		if(fdd < 0 )
		{
			printf("create error\n");
			sceIoClose(fds);
			return -2;
		}

		do
		{
			readed = result = sceIoRead(fds, file_buf , buf_size );
			if(result < 0 )
			{
				printf("read error\n");
				break;
			}
			result = sceIoWrite(fdd, file_buf , result );
			if(result != readed)
			{
				printf("write error\n");
				result = -1;
			}
		}while(result>0);

		sceIoClose(fds);
		sceIoClose(fdd);

		if(result < 0)
			return result;
	}
	else
	{
		printf("Remove %s:",src);
		result = sceIoRemove(src);
		if(result < 0)
		{
			printf("error\n");
			return result;
		}
	}

	printf("OK\n");

	return 0;
}

/****************************************************************************
	copy/remove directry
****************************************************************************/
static int dir_copy_one(const char *src,const char *dst)
{
	char *path_s,*path_d;
	SceUID fd;
	SceIoDirent *dir;
	int result = 0;

	fd = sceIoDopen(src);

	if(fd < 0)
	{
		switch(fd)
		{
		case 0x80010002:
			printf("Directry %s : not found\n",src);
			break;
		default:
			printf("sceIoDopen(%s) : error %08x\n",src,fd);
		}
		return fd;
	}

	printf("---- Directry %s\n",src);
	if(dst)
	{
		printf("Mkdir %s:",dst);
		result = sceIoMkdir(dst,0x777);
		if( result < 0 )
			printf("Err %08X\n",result);
		else
			printf("OK\n");
	}

#if USE_MALLOC
	path_s = (char *)malloc(128+128+sizeof(SceIoDirent));
#else
	int memid = sceKernelAllocPartitionMemory(2,"dir heap",PSP_SMEM_Low, 128+128+sizeof(SceIoDirent), NULL);
	if(memid<0)
	{
		printf("Memory alloc error\n");
		return memid;
	}
	path_s = (char *)sceKernelGetBlockHeadAddr(memid);
#endif
	path_d = dst ? (path_s+128) : NULL;
	dir    = (SceIoDirent *)(path_s+256);

	while(sceIoDread(fd, dir) > 0)
	{
		sprintf(path_s,"%s/%s",src,dir->d_name);
		if(dst)
			sprintf(path_d,"%s/%s",dst,dir->d_name);

		if((dir->d_stat.st_attr & FIO_SO_IFDIR) )
		{
			// directry
			if(dir->d_name[0]!='.')
			{
				result = dir_copy_one(path_s,path_d);
			}
		}
		else
		{
			// file
			result = copy_file(path_s,path_d);
		}
	}
	sceIoDclose(fd);

	//
	if(!dst)
	{
		printf("Rmdir %s:",src);
		result = sceIoRmdir(src);
		if(result < 0)
			printf("err %08X\n",result);
		else
			printf("OK\n");
	}

#if USE_MALLOC
		free(path_s);
#else
		sceKernelFreePartitionMemory(memid);
#endif
	return result;
}

/****************************************************************************
	copy directry
****************************************************************************/
int copy_directry(const char *src,const char *dst)
{
	return dir_copy_one(src,dst);
}

/****************************************************************************
	remove directry
****************************************************************************/
int remove_directry(const char *src)
{
	return dir_copy_one(src,NULL);
}

/****************************************************************************
	make dummy file
****************************************************************************/
int make_dummy_file(const char *path,int size)
{
	SceUID fd;
	int result;

	printf("make dummy file %s:",path);

	fd = sceIoOpen(path, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);
	if(fd < 0 )
	{
		printf("create error %08x\n",fd);
		return fd;
	}

	memset(file_buf,0,sizeof(file_buf));
	while(size)
	{
		int req_size = size > sizeof(file_buf) ? sizeof(file_buf) : size;
		result = sceIoWrite(fd, file_buf , req_size );
		if(result != req_size)
		{
			sceIoClose(fd);
			printf("write error\n");
			return result;
		}
		size -= req_size;
	}
	sceIoClose(fd);
	printf("OK\n");
}

/****************************************************************************
	check directry

	non thread safe function !!
****************************************************************************/
static SceIoDirent chk_dir;

char *check_directry(const char *base_path)
{
	SceUID fd;

	fd = sceIoDopen(base_path);
	if(fd < 0 ) return NULL;

	while(sceIoDread(fd, &chk_dir) > 0)
	{
		if((chk_dir.d_stat.st_attr & FIO_SO_IFDIR) )
		{
			// directry
			if(chk_dir.d_name[0]!='.')
			{
				sceIoDclose(fd);
				return chk_dir.d_name;
			}
		}
	}
	sceIoDclose(fd);
	return NULL;
}

/****************************************************************************
	sunc NAND Flash memory
****************************************************************************/
static int flash_sync()
{
	int result;

	// sync NAND flash
	result = sceIoDevctl("flash0:",0x5802,NULL,0,NULL,0);
	if(result<0)
		printf("flash0: sync error %08X\n",result);

	result = sceIoDevctl("flash1:",0x5802,NULL,0,NULL,0);
	if(result<0)
		printf("flash1: sync error %08X\n",result);

	return result;
}

/****************************************************************************
	enable/protect  flash0 write
****************************************************************************/
static int flash0_enable_write(void)
{
	int result;

	printf("flash0 write enable\n");

	result = sceIoUnassign("flash0:");
	if( result < 0)
	{
		printf("flash0 unassign error %08X\n",result);
	}
	else
	{
		result = sceIoAssign("flash0:", "lflash0:0,0", "flashfat0:", IOASSIGN_RDWR, NULL, 0);
		if( result < 0)
		{
			printf("flash0 assign error %08X\n",result);
		}
	}
	return result;
}

static int flash0_protect(void)
{
	int result;

	printf("flash0 write protect\n");

	flash_sync();

	result = sceIoUnassign("flash0:");
	if( result < 0)
	{
		printf("flash0 unassign error %08X\n",result);
	}
	else
	{
		result = sceIoAssign("flash0:", "lflash0:0,0", "flashfat0:", IOASSIGN_RDONLY, NULL, 0);
		if( result < 0)
		{
			printf("flash0 assign error %08X\n",result);
		}
	}
	return result;
}

/****************************************************************************
ui
****************************************************************************/
#define NOWARNING 0x1000

static int flash_title(const char *title_string,int fs_num)
{
	const char *duplicate_warning;
	const char *start_warning;
	int hold_time;
	char *str;
	int flags = fs_num & 0xfffffff0;

	fs_num &= 0x0f;

	pspDebugScreenClear();
	printf(title_string,src_path,dst_path);
	printf("\n\n");

	// show installed files
	str = check_directry("flash0:/dh/");
	if(str)
		printf(STR_FLASH0_INST_MSG,str);
	else
		printf(STR_FLASH0_NONE_MSG);


	str = check_directry("flash1:/dh/");
	if(str)
		printf(STR_FLASH1_INST_MSG,str);
	else
		printf(STR_FLASH1_NONE_MSG);

	if(flags&NOWARNING)
	{
		// no warning
		duplicate_warning = "";
		start_warning = STR_FLASH_NO_WARNING;
		hold_time = 1*1000;
	}
	else if(fs_num==0)
	{
		duplicate_warning = STR_FLASH_DUP_WARNING;
		start_warning = STR_FLASH0_WARNING;
		hold_time = 5*1000;
	}
	else
	{
		duplicate_warning = STR_FLASH_DUP_WARNING;
		start_warning = STR_FLASH1_WARNING;
		hold_time = 1*1000;
	}

	// duplicate warning 
	printf(duplicate_warning);

	// write warning 
	pspDebugScreenSetTextColor(COLOR_WARNING_FG);
	pspDebugScreenSetBackColor(COLOR_WARNING_BG);
	printf(start_warning);
	pspDebugScreenSetTextColor(COLOR_NORMAL_FG);
	pspDebugScreenSetBackColor(COLOR_NORMAL_BG);

	if( wait_hold_button(PSP_CTRL_CIRCLE,hold_time) == 0)
		return 0;

	// cancel
	printf(STR_FLASH_CANCEL);
	printf(STR_FLASH_DONE);
	wait_key();
	return -1;
}

static int flash_done(void)
{
	printf(STR_FLASH_DONE);
	wait_key();
	return 0;
}

/****************************************************************************
	flash install
****************************************************************************/
static const char dummy_file[] = "flash1:/dh/dummy.dat";

int flash1_install(void)
{
	sprintf(src_path,"%s/flash1",ms_cur_path()    ); // ms
	sprintf(dst_path,"flash1:%s",flash_cur_path() ); // flash

	if(flash_title(STR_FLASH1_WORK,1)) return -1;

#if 0
	// remove registry for TEST!!
	sceIoRemove("flash1:/registry/system.ireg");
	sceIoRemove("flash1:/registry/system.dreg");
#endif

	remove_directry(dst_path);
	make_directry(dst_path);

	// dummy file for keep free space
	if( make_dummy_file(dummy_file,FLASH1_KEEP_FREE_SIZE) >= 0)
		copy_directry(src_path,dst_path);

	// remove dummy file
	printf("remove %s\n",dummy_file);
	sceIoRemove(dummy_file);

	flash_sync();
	flash_done();
	return 0;
}

/****************************************************************************
	flash install
****************************************************************************/
int flash1_backup(void)
{
	sprintf(src_path,"flash1:%s",flash_cur_path() ); // flash
	sprintf(dst_path,"%s/flash1",ms_cur_path()    ); // ms

	if(flash_title(STR_FLASH1_BACKUP,1|NOWARNING)) return -1;

	remove_directry(dst_path);
	make_directry(dst_path);
	copy_directry(src_path,dst_path);
	flash_sync();
	flash_done();

	return 0;
}

/****************************************************************************
	flash install
****************************************************************************/
int flash1_cleanup(void)
{
	src_path[0] = 0;
	strcpy(dst_path,"flash1:/dh");
	if(flash_title(STR_FLASH1_CLEANUP,1)) return -1;

	// remove dest dir
	remove_directry(dst_path);
	flash_sync();

	flash_done();

	return 0;
}
/****************************************************************************
  flash0 font install
****************************************************************************/
int flash0_font(void)
{
	sprintf(src_path,"%s/flash0/font",ms_cur_path()    ); // ms
	sprintf(dst_path,"flash0:%s/font",flash_cur_path() ); // flash

	if(flash_title(STR_FLASH0_FONT,0)) return -1;

	if( flash0_enable_write()>=0)
	{
		remove_directry(dst_path);
		make_directry(dst_path);
		copy_directry(src_path,dst_path);

		flash0_protect();
	}

	flash_done();

	return 0;
}

/****************************************************************************
	flash0 install
****************************************************************************/
int flash0_kd_resource(void)
{
	sprintf(src_path,"%s/flash0/kd/resource",ms_cur_path()    ); // ms
	sprintf(dst_path,"flash0:%s/kd/resource",flash_cur_path() ); // flash

	if(flash_title(STR_FLASH0_KDRES,0)) return -1;

	if( flash0_enable_write()>=0)
	{
		remove_directry(dst_path);
		make_directry(dst_path);
		copy_directry(src_path,dst_path);

		flash0_protect();
	}
	flash_done();
	return 0;
}

/****************************************************************************
****************************************************************************/
int flash0_cleanup(void)
{
	src_path[0] = 0;
	strcpy(dst_path,"flash0:/dh");

	if(flash_title(STR_FLASH0_CLEANUP,0)) return -1;
	if( flash0_enable_write()>=0)
	{
		// remove dest dir
		remove_directry(dst_path);
		flash0_protect();
	}

	flash_done();

	return 0;
}
